package com.ethornium.analysis;

import com.ethornium.encoding.Contract;
import com.ethornium.encoding.InternalRepresentation;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.math.BigInteger;
import org.web3j.crypto.Hash;
import java.util.*;

public class PreAnalysis {
    private static final Logger LOGGER = LogManager.getLogger(PreAnalysis.class);

    private final Map<Integer, InternalRepresentation> internalRepresentation;
    private final Vector<Integer> jumpDests;
    private final Integer maxPc;
    private final String contractID;
    private final Map<Contract, Integer> contracts;
    private Integer pc;
    private Integer next;
    private InternalRepresentation bytecode;
    private ArrayList<BigInteger> stack;
    private final Map<BigInteger, BigInteger> memory;
    private final Map<BigInteger, BigInteger> storage;
    private final Set<Integer> visited;
    int cycle;
    int boundary;
    private String bytecodeName;
    final private boolean debug;
    PreAnalysis(Map<Integer, InternalRepresentation> internalRepresentation,
                 Vector<Integer> jumpDests,
                 Integer maxPc,
                 final int contractID,
                 Map<Contract, Integer> contracts,
                 int boundary, boolean debug){
         this.internalRepresentation = internalRepresentation;
         this.jumpDests = jumpDests;
         this.maxPc = maxPc;
         this.contractID = Integer.toString(contractID);
         this.contracts = contracts;
         this.bytecode = null;
         this.pc = null;
         this.next = null;
         this.stack = new ArrayList<>();
         this.memory = new HashMap<>();
         this.storage = new HashMap<>();
         this.cycle = 0;
         this.visited = new HashSet<>();
         this.boundary = boundary;
         bytecodeName = "";
         this.debug = debug;
    }
    private BigInteger getStackEl(int index){
        if ((index < 0) || (index >= stack.size())){
            LOGGER.error("Getting stack element failed!");
            return null;
        }
        else{
            return stack.get(index);
        }
    }
    private void deleteStackEl(int index){
        if (!((index < 0) || (index >= stack.size()))){
            stack.remove(index);
        }
        else{
            LOGGER.error("Removing stack element failed!");
        }
    }
    private void setStackEl(int index, BigInteger value){
        if (!((index < 0) || (index >= stack.size()))){
            stack.set(index, value);
        } else{
            LOGGER.error("Setting stack element failed!");
        }
    }
    public void process(int pc) {
        if (!visited.isEmpty()) {
            if (visited.contains(pc)) {
                if (debug)
                    LOGGER.debug("Not entering the cycles");
                return;
            }
        }
        //if (cycle > boundary) return;
        visited.add(pc);
        this.pc = pc;
        bytecode = internalRepresentation.get(pc);
        if (bytecode == null) return;
        bytecode.setVisited();
        caller();
    }

    private void putComputedArgument(BigInteger arg){
        if (arg != null){
            bytecode.putComputedArgument(arg.toString());
        }
        else{
            bytecode.putComputedArgument(null);
        }
    }

    private void putComputedArgument(Integer arg){
        if (arg != null){
            bytecode.putComputedArgument(arg.toString());
        }
        else{
            bytecode.putComputedArgument(null);
        }
    }

    private void overapproxBinary() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        BigInteger rez = null;
        stack.add(rez);
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void overApproxAppendToStack(){
        BigInteger rez = null;
        stack.add(rez);
        next = getNext();
        log(valToString(rez), valToString(next));
        process(next);
    }

    private void STOP0() {
        valToString(next);
    }

    private void ADD1() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            rez = x.add(y);
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void MUL2() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            rez = x.multiply(y);
            stack.add(rez);
        }
        else{
            if ((x != null) && (x.compareTo(BigInteger.valueOf(0)) == 0)){
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
            else{
                if ((y != null) && (y.compareTo(BigInteger.valueOf(0)) == 0)){
                    rez = BigInteger.valueOf(0);
                    stack.add(rez);
                }
                else{
                    stack.add(null);
                }
            }
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void SUB3() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            rez = x.subtract(y);
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void log(String... st){
        String s = "";
        for (String s1: st){
            s = s + " " + s1;
        }
        if (debug) {
            LOGGER.info(Integer.toString(pc) + " " + bytecodeName + " " + s + " Stack Size:" + Integer.toString(stack.size()));
        }
    }

    private String valToString(Integer val){
        String rez = "null";
        if (val != null){
            return val.toString();
        }
        return rez;
    }

    private String valToString(BigInteger val){
        String rez = "null";
        if (val != null){
            return val.toString();
        }
        return rez;
    }

    private void DIV4() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            if (y.compareTo(BigInteger.valueOf(0)) == 0){
                stack.add(BigInteger.valueOf(0));
                rez = BigInteger.valueOf(0);
            }
            else{
                rez = x.divide(y);
                stack.add(rez);
            }
        }
        else{
            if ((x != null) && (x.compareTo(BigInteger.valueOf(0)) == 0)){
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
            else{
                if ((y != null) && (y.compareTo(BigInteger.valueOf(0)) == 0)){
                    rez = BigInteger.valueOf(0);
                    stack.add(rez);
                }
                else{
                    stack.add(rez);
                }
            }
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void SDIV5() {
        overapproxBinary();
    }

    private void MOD6() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            if (y.compareTo(BigInteger.valueOf(0)) == 0){
                stack.add(BigInteger.valueOf(0));
                rez = BigInteger.valueOf(0);
            }
            else{
                rez = x.mod(y);
                stack.add(rez);
            }
        }
        else{
            if ((x != null) && (x.compareTo(BigInteger.valueOf(0)) == 0)){
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
            else{
                if ((y != null) && (y.compareTo(BigInteger.valueOf(0)) == 0)){
                    rez = BigInteger.valueOf(0);
                    stack.add(rez);
                }
                else{
                    stack.add(rez);
                }
            }
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void SMOD7() {
        overapproxBinary();
    }

    private void ADDMOD8() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger z = getStackEl(stack.size() - 3);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null && z != null) {
            rez = (x.add(y)).mod(z);
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(z);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(x), valToString(rez), valToString(next));
        process(next);
    }
    private void MULMOD9() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger z = getStackEl(stack.size() - 3);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null && z != null) {
            rez = (x.multiply(y)).mod(z);
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(z);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(x), valToString(rez), valToString(next));
        process(next);
    }
    private void EXP0a() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            Integer pow = y.intValue();
            rez = x.pow(pow);
            stack.add(rez);
        }
        else{
            stack.add(null);
            stack.add(rez);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }
    private void SIGNEDEXTEND0b() {
        overapproxBinary();
    }
    private void LT10() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        BigInteger rez;
        if (x!= null && y != null) {
            if (x.compareTo(y) == -1) {
                rez = BigInteger.valueOf(1);
                stack.add(rez);
            } else {
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
        }
        else{
            rez = null;
            stack.add(rez);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }
    private void GT11() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        BigInteger rez;
        if (x!= null && y != null) {
            if (x.compareTo(y) == 1) {
                rez = BigInteger.valueOf(1);
                stack.add(rez);
            } else {
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
        }
        else{
            rez = null;
            stack.add(rez);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }
    private void SLT18() {
        overapproxBinary();
    }
    private void SGT19() {
        overapproxBinary();
    }
    private void EQ14() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        BigInteger rez;
        if (x!= null && y != null) {
            if (x.compareTo(y) == 0) {
                rez = BigInteger.valueOf(1);
                stack.add(rez);
            } else {
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
        }
        else{
            rez = null;
            stack.add(rez);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }
    private void ISZERO15() {
        BigInteger x = getStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        BigInteger rez = null;
        if (x!= null) {
            if (x.compareTo(BigInteger.valueOf(0)) == 0) {
                rez = BigInteger.valueOf(1);
            } else {
                rez = BigInteger.valueOf(0);
            }
        }
        stack.add(rez);
        putComputedArgument(x);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(rez), valToString(next));
        process(next);
    }
    private void AND16() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            rez = x.and(y);
            stack.add(rez);
        }
        else{
            if ((x != null) && (x.compareTo(BigInteger.valueOf(0)) == 0)){
                rez = BigInteger.valueOf(0);
                stack.add(rez);
            }
            else{
                if ((y != null) && (y.compareTo(BigInteger.valueOf(0)) == 0)){
                    rez = BigInteger.valueOf(0);
                    stack.add(rez);
                }
                else{
                    stack.add(null);
                }
            }
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }
    private void OR17() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            rez = x.or(y);
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }
    private void XOR18() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger y = getStackEl(stack.size() - 2);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        if (x!= null && y != null) {
            rez = x.xor(y);
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(y);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(y), valToString(rez), valToString(next));
        process(next);
    }

    private void NOT19() {
        BigInteger x = getStackEl(stack.size() - 1);
        BigInteger rez = null;
        deleteStackEl(stack.size() - 1);
        if (x!= null) {
            rez = x.not();
            stack.add(rez);
        }
        else{
            stack.add(null);
        }
        putComputedArgument(x);
        putComputedArgument(rez);
        next = getNext();
        log(valToString(x), valToString(rez), valToString(next));
        process(next);
    }
    private void BYTE1a() {
        overapproxBinary();
    }
    //TODO: check
    private void SHA320() {
        BigInteger pos = getStackEl(stack.size() - 1);
        BigInteger s = getStackEl(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        putComputedArgument(pos);
        putComputedArgument(s);
        BigInteger rez = null;

        ArrayList<Byte> byteList = new ArrayList();
        if ((pos !=null) && (s !=null)){
            int posInt = pos.intValue();
            int sInt = s.intValue();
            for (int ind = posInt; ind < posInt + sInt; ind++){
                BigInteger val = memory.get(ind);
                if (val != null){
                    byte[] valAsByteArray = val.toByteArray();
                    for (byte b: valAsByteArray){
                        byteList.add(b);
                    }
                }
                else {
                    byte byte0 = BigInteger.ZERO.byteValueExact();
                    byteList.add(byte0);
                }
            }
            byte[] byteArray = new byte[byteList.size()];
            for(int i = 0; i < byteList.size(); i++) {
                byteArray[i] = byteList.get(i).byteValue();
            }
            byte[] rezByte = Hash.sha3(byteArray);
            rez = new BigInteger(1, rezByte);
        }

        putComputedArgument(rez);
        stack.add(rez);
        next = getNext();
        log(valToString(pos), valToString(s), valToString(rez), valToString(next));
        process(next);
    }
    private void ADDRESS30() {
        overApproxAppendToStack();
    }
    private void BALANCE31(){
        overApproxAppendToStack();
    }
    private void ORIGIN32(){
        overApproxAppendToStack();
    }
    private void CALLER33() {
        overApproxAppendToStack();
    }
    private void CALLVALUE34() {
        overApproxAppendToStack();
    }
    private void CALLDATALOAD35() {
        BigInteger pos = getStackEl(stack.size() - 1);
        putComputedArgument(pos);
        BigInteger rez = null;
        /*if (pos.compareTo(BigInteger.valueOf(0)) == 0){
            rez = BigInteger.valueOf(0);
            rezz = rez.toString();
        }*/
        setStackEl(stack.size() -1, rez);
        next = getNext();
        log(valToString(pos), valToString(rez), valToString(next));
        process(next);
    }
    private void CALLDATASIZE36() {
        overApproxAppendToStack();
    }
    private void CALLDATACOPY37() {
        overApproxAppendToStack();
    }
    private void CODESIZE38() {
        BigInteger rez = BigInteger.valueOf(maxPc + 1);
        stack.add(rez);
        next = getNext();
        log(valToString(rez), valToString(next));
        process(next);
    }
    private void CODECOPY39() {
        BigInteger posMem = getStackEl(stack.size() - 1);
        BigInteger posCom = getStackEl(stack.size() - 2);
        BigInteger s = stack.get(stack.size() - 3);
        putComputedArgument(posMem);
        putComputedArgument(posCom);
        putComputedArgument(s);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        memory.put(null, null);
        log(valToString(posMem), valToString(posCom), valToString(s), valToString(next));
        next = getNext();
        process(next);
    }
    private void GASPRICE3a(){
        overApproxAppendToStack();
    }
    private void EXTCODESIZE3b() {
        overApproxAppendToStack();
    }
    private void EXTCODECOPY3c() {
        overApproxAppendToStack();
    }
    private void BLOCKHASH40() {
        overApproxAppendToStack();
    }
    private void COINBASE41(){
        overApproxAppendToStack();
    }
    private void TIMESTAMP42(){
        overApproxAppendToStack();
    }
    private void NUMBER43(){
        overApproxAppendToStack();
    }
    private void DIFFICULTY44(){
        overApproxAppendToStack();
    }
    private void GASLIMIT45(){
        overApproxAppendToStack();
    }
    private void POP50() {
        deleteStackEl(stack.size() - 1);
        next = getNext();
        log(valToString(next));
        process(next);
    }

    private void MLOAD51() {
        Boolean found = false;
        BigInteger bigMa = getStackEl(stack.size() - 1);
        putComputedArgument(bigMa);
        Integer ma;
        BigInteger value = null;
        if (bigMa != null) {
            ma = bigMa.intValue();

            if (memory.size() > 0) {
                if (memory.keySet().contains(null)) {

                } else {
                    if (ma % 32 == 0) {
                        value = memory.get(ma);
                        found = true;
                        putComputedArgument(ma);
                        if (value != null) {
                            putComputedArgument(value);
                            setStackEl(stack.size() - 1, value);
                        } else {
                            setStackEl(stack.size() - 1, BigInteger.valueOf(0));
                            putComputedArgument(BigInteger.valueOf(0));
                        }
                        log(valToString(ma), valToString(value), Integer.toString(next));
                    }

                }
            }
        }

        if (!found){
            log(valToString(bigMa), valToString(value), Integer.toString(next));
            setStackEl(stack.size() - 1, null);
        }
        next = getNext();
        process(next);
    }
    private void MSTORE52() {
        BigInteger bigMa = getStackEl(stack.size() - 1);
        BigInteger ma = null;
        if (bigMa !=  null)
            ma = stack.get(stack.size() - 1);
        BigInteger v = stack.get(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        putComputedArgument(ma);
        putComputedArgument(v);
        if ((ma != null) && (ma.mod(BigInteger.valueOf(32)) == BigInteger.ZERO)){
            memory.put(ma, v);
        }
        else{
            memory.put(null, v);
        }
        next = getNext();
        log(valToString(ma), valToString(v), Integer.toString(next));
        process(next);
    }
    private void MSTORE853() {
        Integer v = null ,ma = null;
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        putComputedArgument(ma);
        putComputedArgument(v);
        next = getNext();
        log(valToString(ma), valToString(v), Integer.toString(next));
        process(next);
    }
    private void SLOAD54() {
        BigInteger ma;
        BigInteger v;
        if (getStackEl(stack.size() - 1) != null) {
            ma = stack.get(stack.size() - 1);
            v = storage.get(ma);
        }
        else{
            ma = null;
            v = null;
        }
        setStackEl(stack.size() - 1, v);
        putComputedArgument(ma);
        putComputedArgument(v);
        next = getNext();
        log(valToString(ma), valToString(v), valToString(next));
        process(next);
    }

    private void JUMP56() {
        BigInteger i;
        if (stack.size() - 1 < 0) {
            i = null;
        }
        else {
            i = stack.get(stack.size() - 1);
            deleteStackEl(stack.size() - 1);
        }
        Integer jump = null;
        if (i != null){
            jump = i.intValue();
        }
        else{
            if (debug) {
                LOGGER.error("Pre-analysis cannot find a jump destination at: " + next.toString());
            }
        }
        putComputedArgument(jump);
        if (jump != null) {
            log(Integer.toString(jump));
        }
        else{
            log("null");
        }
        if (jump != null) {
            process(jump);
        }
        else{
            next = getNext();
            process(next);
        }
    }

    private void JUMPI57() {
        BigInteger i = getStackEl(stack.size() - 1);
        Integer jump = null;
        if (i != null){
            jump = i.intValue();
        }
        else{
            if (debug) {
                LOGGER.error("Pre-analysis cannot find a jump destination at: " + next.toString());
            }
        }
        putComputedArgument(jump);
        BigInteger bcond = stack.get(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        Integer cond;
        if (bcond != null){
            cond = bcond.intValue();
        }
        else{
            cond = null;
        }
        putComputedArgument(cond);
        next = getNext();
        log(valToString(cond), valToString(jump), Integer.toString(next));
        if (cond != null){
            if (cond != 0){
                process(jump);
            }
            else{
                process(next);
            }
        }
        else{
            if (jump!=null) {
                ArrayList<BigInteger> stackSave = stack;
                int nextSave = next;
                process(jump);
                next = nextSave;
                cycle = 0;
                stack = stackSave;
                process(nextSave);
            }
            else{
                process(next);
            }
        }
    }

    private void JUMPDEST5b() {
        next = getNext();
        log(Integer.toString(next));
        process(next);
    }
    private void PC58(){
        overApproxAppendToStack();
    }
    private void MSIZE59(){
        overApproxAppendToStack();
    }
    private void GAS90(){
        overApproxAppendToStack();
    }
    private int getNext(){
        Iterator<Map.Entry<Integer, InternalRepresentation>> it = internalRepresentation.entrySet().iterator();
        int min = -1;
        boolean found = false;
        while (it.hasNext()) {
            Map.Entry <Integer, InternalRepresentation> first = it.next();
            if (first.getKey() > pc){
                if (!found) {
                    found = true;
                    min = first.getKey();
                }
                else{
                    if (first.getKey() < min)
                        min = first.getKey();
                }

            }
        }
        return min;
    }
    
    private void PUSH160(int x) {
        BigInteger arg = new BigInteger(bytecode.getArg(), 16);
        stack.add(arg);
        next = getNext();
        log(valToString(arg), valToString(next));
        process(next);
    }
    private void DUPS80(int x) {
        BigInteger arg = null;
        if (stack.size() >= x) {
            arg = stack.get(stack.size() - x);
        }
        stack.add(arg);
        putComputedArgument(arg);
        next = getNext();
        log(valToString(arg), valToString(next));
        process(next);
    }
    private void SWAPS90(int x) {
        BigInteger arg1 = null;
        if (stack.size() > 0) {
            arg1 = stack.get(stack.size() - 1);
        }
        putComputedArgument(arg1);
        BigInteger arg2;
        if (stack.size() - x - 1 < 0) {
            arg2 =  null;
        } else {
            arg2 = stack.get(stack.size() - x - 1);
        }
        putComputedArgument(arg2);
        setStackEl(stack.size() - 1, arg2);
        if (stack.size() - x - 1 < 0) {
            if (debug) {
                LOGGER.error("ERROR performaing SWAP at " + pc);
            }
        }
        else{
            setStackEl(stack.size() - x - 1, arg1);
        }
        next = getNext();
        log(valToString(arg1), valToString(arg2), valToString(next));
        process(next);
    }
    private void SSTORE55() {
        BigInteger ma;
        BigInteger v;
        if (getStackEl(stack.size() - 1) != null){
            ma = stack.get(stack.size() - 1);
            v = stack.get(stack.size() - 2);
        }
        else{
            ma = null;
            v = null;
        }
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        putComputedArgument(ma);
        putComputedArgument(v);
        storage.put(ma, v);
        next = getNext();
        log(valToString(ma), valToString(v), valToString(next));
        process(next);
    }

    private void LOGSa0(int x) {
        /*Integer pos = null;
        if (stack.get(stack.size() - 1) != null) {
            pos = stack.get(stack.size() - 1).intValue();
        }
        BigInteger s = stack.get(stack.size() - 2);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        putComputedArgument(pos);
        putComputedArgument(s);
        for (int i = 1; i<= x; i ++ ){
            deleteStackEl(stack.size() - 1);
        }*/
        next = getNext();
        log(valToString(next));
        process(next);
    }

    private void CREATEC(){
        CALL();
    }

    private void CALL() {
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        deleteStackEl(stack.size() - 1);
        stack.add(BigInteger.valueOf(1));
        memory.put(null, null);
        next = getNext();
        log(valToString(1), valToString(next));
        process(next);
    }

    private void CALLCODE(){
        CALL();
    }
    private void RETURN(){
        log(valToString(next));
    }
    private  void DELEGATECALL(){
        CALL();
    }
    private void SELFDESTRUCT(){
        log(valToString(next));
    }
    private void REVERT() {
        log(valToString(next));
    }
    private void NOOPff() {
        next = getNext();
        log(valToString(next));
        process(next);
    }


    
    private void caller(){
        switch (bytecode.getOpCode()) {
            case 0:
                bytecodeName = "STOP";
                STOP0();
                break;
            case 1:
                bytecodeName = "ADD";
                ADD1();
                break;
            case 2:
                bytecodeName = "MUL";
                MUL2();
                break;
            case 3:
                bytecodeName = "SUB";
                SUB3();
                break;
            case 4:
                bytecodeName = "DIV";
                DIV4();
                break;
            case 5:
                bytecodeName = "SDIV";
                SDIV5();
                break;
            case 6:
                bytecodeName = "MOD";
                MOD6();
                break;
            case 7:
                bytecodeName = "SMOD";
                SMOD7();
                break;
            case 8:
                bytecodeName = "ADDMOD";
                ADDMOD8();
                break;
            case 9:
                bytecodeName = "MULMOD";
                MULMOD9();
                break;
            case 10:
                bytecodeName = "EXP";
                EXP0a();
                break;
            case 11:
                bytecodeName = "SIGNEDEXTEND";
                SIGNEDEXTEND0b();
                break;
            case 16:
                bytecodeName = "LT";
                LT10();
                break;
            case 17:
                bytecodeName = "GT";
                GT11();
                break;
            case 18:
                bytecodeName = "SLT";
                SLT18();
                break;
            case 19:
                bytecodeName = "SGT";
                SGT19();
                break;
            case 20:
                bytecodeName = "EQ";
                EQ14();
                break;
            case 21:
                bytecodeName = "ISZ";
                ISZERO15();
                break;
            case 22:
                bytecodeName = "AND";
                AND16();
                break;
            case 23:
                bytecodeName = "OR";
                OR17();
                break;
            case 24:
                bytecodeName = "XOR";
                XOR18();
                break;
            case 25:
                bytecodeName = "NOT";
                NOT19();
                break;
            case 26:
                bytecodeName = "BYTE";
                BYTE1a();
                break;
            case 32:
                bytecodeName = "SHA";
                SHA320();
                break;
            case 48:
                bytecodeName = "ADDRESS";
                ADDRESS30();
                break;
            case 49:
                bytecodeName = "BALANCE";
                BALANCE31();
                break;
            case 50:
                bytecodeName = "ORIGIN";
                ORIGIN32();
                break;
            case 51:
                bytecodeName = "CALLER";
                CALLER33();
                break;
            case 52:
                bytecodeName = "CALLVALUE";
                CALLVALUE34();
                break;
            case 53:
                bytecodeName = "CALLDATALOAD";
                CALLDATALOAD35();
                break;
            case 54:
                bytecodeName = "CALLDATASIZE";
                CALLDATASIZE36();
            case 55:
                bytecodeName = "CALLDATACOPY";
                CALLDATACOPY37();
            case 56:
                bytecodeName = "CODESIZE";
                CODESIZE38();
                break;
            case 57:
                bytecodeName = "CODECOPY";
                CODECOPY39();
                break;
            case 58:
                bytecodeName = "GASPRICE";
                GASPRICE3a();
                break;
            case 59:
                bytecodeName = "EXTCODESIZE";
                EXTCODESIZE3b();
                break;
            case 60:
                bytecodeName = "EXTCODECOPY";
                EXTCODECOPY3c();
                break;
            case 64:
                bytecodeName = "BLOCKHASH";
                BLOCKHASH40();
                break;
            case 65:
                bytecodeName = "COINBASE";
                COINBASE41();
                break;
            case 66:
                bytecodeName = "TIMESTAMP";
                TIMESTAMP42();
                break;
            case 67:
                bytecodeName = "NUMBER";
                NUMBER43();
                break;
            case 68:
                bytecodeName = "DIFFICULTY";
                DIFFICULTY44();
                break;
            case 69:
                bytecodeName = "GASLIMIT";
                GASLIMIT45();
                break;
            case 80:
                bytecodeName = "POP";
                POP50();
                break;
            case 81:
                bytecodeName = "MLOAD";
                MLOAD51();
                break;
            case 82:
                bytecodeName = "MSTORE";
                MSTORE52();
                break;
            case 83:
                bytecodeName = "MSTORE8";
                MSTORE853();
                break;
            case 84:
                bytecodeName = "SLOAD";
                SLOAD54();
                break;
            case 85:
                bytecodeName = "SSTORE";
                SSTORE55();
                break;
            case 86:
                bytecodeName = "JUMP";
                JUMP56();
                break;
            case 87:
                bytecodeName = "JUMPI";
                JUMPI57();
                break;
            case 88:
                bytecodeName = "PC";
                PC58();
                break;
            case 89:
                bytecodeName = "MSIZE";
                MSIZE59();
                break;
            case 90:
                bytecodeName = "GAS";
                GAS90();
                break;
            case 91:
                bytecodeName = "JUMPDEST";
                JUMPDEST5b();
                break;
            case 96:
                bytecodeName = "PUSH";
                PUSH160(1);
                break;
            case 97:
                bytecodeName = "PUSH";
                PUSH160(2);
                break;
            case 98:
                bytecodeName = "PUSH";
                PUSH160(3);
                break;
            case 99:
                bytecodeName = "PUSH";
                PUSH160(4);
                break;
            case 100:
                bytecodeName = "PUSH";
                PUSH160(5);
                break;
            case 101:
                bytecodeName = "PUSH";
                PUSH160(6);
                break;
            case 102:
                bytecodeName = "PUSH";
                PUSH160(7);
                break;
            case 103:
                bytecodeName = "PUSH";
                PUSH160(8);
                break;
            case 104:
                bytecodeName = "PUSH";
                PUSH160(9);
                break;
            case 105:
                bytecodeName = "PUSH";
                PUSH160(10);
                break;
            case 106:
                bytecodeName = "PUSH";
                PUSH160(11);
                break;
            case 107:
                bytecodeName = "PUSH";
                PUSH160(12);
                break;
            case 108:
                bytecodeName = "PUSH";
                PUSH160(13);
                break;
            case 109:
                bytecodeName = "PUSH";
                PUSH160(14);
                break;
            case 110:
                bytecodeName = "PUSH";
                PUSH160(15);
                break;
            case 111:
                bytecodeName = "PUSH";
                PUSH160(16);
                break;
            case 112:
                bytecodeName = "PUSH";
                PUSH160(17);
                break;
            case 113:
                bytecodeName = "PUSH";
                PUSH160(18);
                break;
            case 114:
                bytecodeName = "PUSH";
                PUSH160(19);
                break;
            case 115:
                bytecodeName = "PUSH";
                PUSH160(20);
                break;
            case 116:
                bytecodeName = "PUSH";
                PUSH160(21);
                break;
            case 117:
                bytecodeName = "PUSH";
                PUSH160(22);
                break;
            case 118:
                bytecodeName = "PUSH";
                PUSH160(23);
                break;
            case 119:
                bytecodeName = "PUSH";
                PUSH160(24);
                break;
            case 120:
                bytecodeName = "PUSH";
                PUSH160(25);
                break;
            case 121:
                bytecodeName = "PUSH";
                PUSH160(26);
                break;
            case 122:
                bytecodeName = "PUSH";
                PUSH160(27);
                break;
            case 123:
                bytecodeName = "PUSH";
                PUSH160(28);
                break;
            case 124:
                bytecodeName = "PUSH";
                PUSH160(29);
                break;
            case 125:
                bytecodeName = "PUSH";
                PUSH160(30);
                break;
            case 126:
                bytecodeName = "PUSH";
                PUSH160(31);
                break;
            case 127:
                bytecodeName = "PUSH";
                PUSH160(32);
                break;
            case 128:
                bytecodeName = "DUP";
                DUPS80(1);
                break;
            case 129:
                bytecodeName = "DUP";
                DUPS80(2);
                break;
            case 130:
                bytecodeName = "DUP";
                DUPS80(3);
                break;
            case 131:
                bytecodeName = "DUP";
                DUPS80(4);
                break;
            case 132:
                bytecodeName = "DUP";
                DUPS80(5);
                break;
            case 133:
                bytecodeName = "DUP";
                DUPS80(6);
                break;
            case 134:
                bytecodeName = "DUP";
                DUPS80(7);
                break;
            case 135:
                bytecodeName = "DUP";
                DUPS80(8);
                break;
            case 136:
                bytecodeName = "DUP";
                DUPS80(9);
                break;
            case 137:
                bytecodeName = "DUP";
                DUPS80(10);
                break;
            case 138:
                bytecodeName = "DUP";
                DUPS80(11);
                break;
            case 139:
                bytecodeName = "DUP";
                DUPS80(12);
                break;
            case 140:
                bytecodeName = "DUP";
                DUPS80(13);
                break;
            case 141:
                bytecodeName = "DUP";
                DUPS80(14);
                break;
            case 142:
                bytecodeName = "DUP";
                DUPS80(15);
                break;
            case 143:
                bytecodeName = "DUP";
                DUPS80(16);
                break;
            case 144:
                bytecodeName = "SWAP";
                SWAPS90(1);
                break;
            case 145:
                bytecodeName = "SWAP";
                SWAPS90(2);
                break;
            case 146:
                bytecodeName = "SWAP";
                SWAPS90(3);
                break;
            case 147:
                bytecodeName = "SWAP";
                SWAPS90(4);
                break;
            case 148:
                bytecodeName = "SWAP";
                SWAPS90(5);
                break;
            case 149:
                bytecodeName = "SWAP";
                SWAPS90(6);
                break;
            case 150:
                bytecodeName = "SWAP";
                SWAPS90(7);
                break;
            case 151:
                bytecodeName = "SWAP";
                SWAPS90(8);
                break;
            case 152:
                bytecodeName = "SWAP";
                SWAPS90(9);
                break;
            case 153:
                bytecodeName = "SWAP";
                SWAPS90(10);
                break;
            case 154:
                bytecodeName = "SWAP";
                SWAPS90(11);
                break;
            case 155:
                bytecodeName = "SWAP";
                SWAPS90(12);
                break;
            case 156:
                bytecodeName = "SWAP";
                SWAPS90(13);
                break;
            case 157:
                bytecodeName = "SWAP";
                SWAPS90(14);
                break;
            case 158:
                bytecodeName = "SWAP";
                SWAPS90(15);
                break;
            case 159:
                bytecodeName = "SWAP";
                SWAPS90(16);
                break;
            case 160:
                bytecodeName = "LOG";
                LOGSa0(0);
                break;
            case 161:
                bytecodeName = "LOG";
                LOGSa0(1);
                break;
            case 162:
                bytecodeName = "LOG";
                LOGSa0(2);
                break;
            case 163:
                bytecodeName = "LOG";
                LOGSa0(3);
                break;
            case 164:
                bytecodeName = "LOG";
                LOGSa0(4);
                break;
            case 240:
                bytecodeName = "CREATE";
                CREATEC();
                break;
            case 241:
                bytecodeName = "CALL";
                CALL();
                break;
            case 242:
                bytecodeName = "CALLCODE";
                CALLCODE();
                break;
            case 243:
                bytecodeName = "RETURN";
                RETURN();
                break;
            case 244:
                bytecodeName = "DELEGATECALL";
                DELEGATECALL();
                break;
            case 253:
                bytecodeName = "REVERT";
                REVERT();
                break;
            //TODO: create special cases for invalid and revert
            case 254:
                bytecodeName = "INVALID";
                REVERT(); //INVALID
                break;
            case 255:
                bytecodeName = "SELFTDESTRUCT";
                SELFDESTRUCT();
                break;
            default:
                bytecodeName = "NOP";
                NOOPff();
                break;
        }
    }
}
