package com.ethornium.z3;

import com.ethornium.encoding.EExpr;
import com.ethornium.encoding.EFunc;
import com.ethornium.encoding.enums.ESort;
import com.microsoft.z3.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.util.ArrayList;

public class Encoder {
    private final Context ctx;
    private final Engine engine;
    private final Var var;
    private final IntSort intSort;
    private final BoolSort boolSort;
    private final ArraySort arraySortInt;
    private final ArraySort arraySortBool;
    private final ArrayList<BoolExpr> rules;
    private final ArrayList<FuncDecl> funcs;
    private final int GUARD;
    private static final Logger LOGGER = LogManager.getLogger(Encoder.class);

    public Encoder(final Context ctx, final Engine engine, final Var var, final IntSort intSort, final BoolSort boolSort,
                   final ArraySort arraySortInt, final ArraySort arraySortBool,
                   ArrayList<BoolExpr> rules, ArrayList<FuncDecl> funcs,
                   final int GUARD) {
        this.ctx = ctx;
        this.engine = engine;
        this.var = var;
        this.intSort = intSort;
        this.boolSort = boolSort;
        this.rules = rules;
        this.funcs = funcs;
        this.arraySortInt = arraySortInt;
        this.arraySortBool = arraySortBool;
        this.GUARD = GUARD;
    }

    private FuncDecl addFunc(EFunc efunc) {
        String funcName = efunc.getName();
        ArrayList<Sort> domainsList = new ArrayList<>();
        for (ESort esort : efunc.getSorts()) {
            switch (esort) {
                case IntSort:
                    domainsList.add(intSort);
                    break;
                case BoolSort:
                    domainsList.add(boolSort);
                    break;
                case ArraySortInt:
                    domainsList.add(arraySortInt);
                    break;
                case ArraySortT:
                    domainsList.add(arraySortBool);
                    break;
                case ArraySortS:
                    domainsList.add(arraySortBool);
                    break;
                case IntConst:
                    domainsList.add(intSort);
                    break;
                case Taint:
                    domainsList.add(boolSort);
                    break;
                case Symbolic:
                    domainsList.add(boolSort);
                    break;
            }
        }
        Sort[] domains = domainsList.toArray(new Sort[domainsList.size()]);
        FuncDecl f = ctx.mkFuncDecl(funcName, domains, boolSort);
        funcs.add(f);
        return f;
    }

    private BoolExpr encodePredicate(EExpr e) {
        FuncDecl f = addFunc(e.getFunction());
        Expr[] arg = encodeArg(e);
        for (Expr arg_el: arg) {
            if (arg_el == null) {
                throw new NullPointerException();
            }
        }
        BoolExpr res = (BoolExpr) f.apply(arg);
        return res;
    }

    private Expr[] encodeArg(EExpr e) {

        ArrayList<Expr> exprsList = new ArrayList<>();
        for (EExpr arg : e.getArg()) {
            exprsList.add(encodeElement(arg));
        }
        Expr[] exprs = exprsList.toArray(new Expr[exprsList.size()]);
        return exprs;
    }

    private Expr encodeElement(EExpr e) {
        Expr[] eargs;
        ArrayList<BoolExpr> arg;
        switch (e.getExpressionType()) {
            case pred:
                return encodePredicate(e);
            case and:
                eargs = encodeArg(e);
                arg = new ArrayList<>();
                for (Expr earg : eargs) {
                    arg.add((BoolExpr) earg);
                }
                return engine.and(arg.toArray(new BoolExpr[arg.size()]));
            case or:
                eargs = encodeArg(e);
                arg = new ArrayList<>();
                for (Expr earg : eargs) {
                    arg.add((BoolExpr) earg);
                }
                return engine.or(arg.toArray(new BoolExpr[arg.size()]));
            case not:
                return engine.not((BoolExpr) encodeElement(e.getArg().get(0)));
            case eq:
                return engine.eq(encodeElement(e.getArg().get(0)), encodeElement(e.getArg().get(1)));
            case implies:
                return engine.implies((BoolExpr) encodeElement(e.getArg().get(0)), (BoolExpr) encodeElement(e.getArg().get(1)));
            case add:
                return engine.add((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case div:
                return engine.div((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case mul:
                /*if  (e.getArg().get(1).getSort() == ESort.IntConst){
                    return engine.mulConst((IntExpr) encodeElement(e.getArg().get(0)), e.getArg().get(1).getIntValue());
            }*/
                return engine.mul((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            /*case power:
                return engine.power((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));*/
            case sub:
                return engine.sub((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case mod:
                return engine.mod((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case selectInt:
                if (encodeElement(e.getArg().get(1)) instanceof ArithExpr) {
                    return engine.selectInt((ArrayExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
                } else {
                    return engine.selectInt((ArrayExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
                }
            case storeInt:
            case storeT:
            case storeS:
                return engine.store((ArrayExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)),
                        encodeElement(e.getArg().get(2)));
            case selectT:
            case selectS:
                if (encodeElement(e.getArg().get(1)) instanceof ArithExpr) {
                    return engine.selectBool((ArrayExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
                } else {
                    return engine.selectBool((ArrayExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
                }
            case prim:
                switch (e.getSort()) {
                    case IntConst:
                        if (e.getStringValue() == null) {
                            return engine.num(e.getIntValue());
                        } else {
                            return engine.num(e.getStringValue());
                        }
                    case IntSort:
                        if (e.getIntValue() >= GUARD) {
                            return var.getVal(e.getIntValue());
                        } else {
                            return var.getVarInt(e.getIntValue());
                        }
                    case ArraySortInt:
                        return var.getVarArr(e.getIntValue());
                    case ArraySortT:
                        return var.getValTaintArr(e.getIntValue());
                    case ArraySortS:
                        return var.getValSymbolArr(e.getIntValue());
                    case Taint:
                        if (e.getIntValue() >= GUARD) {
                            return var.getValT(e.getIntValue());
                        } else {
                            return var.getVarTaint(e.getIntValue());
                        }
                    case Symbolic:
                        if (e.getIntValue() >= GUARD) {
                            return var.getValS(e.getIntValue());
                        } else {
                            return var.getVarSymbol(e.getIntValue());
                        }
                    case BoolConst:
                        if (e.getBoolValue()){
                            return engine.tru();
                        }
                        else{
                            return engine.fal();
                        }
                }
            /*case checkValid:
                return engine.checkValid((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));*/
            case ite:
                return engine.ite((BoolExpr) encodeElement(e.getArg().get(0)), encodeElement(e.getArg().get(1)), encodeElement(e.getArg().get(2)));
            case log256Floor:
                return engine.log256Floor((IntExpr) encodeElement(e.getArg().get(0)));
            case gt:
                return engine.gt((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case le:
                return engine.le((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case lt:
                return engine.lt((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case ge:
                return engine.ge((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case andBitwise:
                return engine.andBitwise((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case orBitwise:
                return engine.orBitwise((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case xorBitwise:
                return engine.xorBitwise((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case notBitwise:
                return engine.notBitwise((IntExpr) encodeElement(e.getArg().get(0)));
            case max:
                return engine.max((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case min:
                return engine.min((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case divCeiling:
                return engine.divCeiling((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case cMem:
                return engine.cMem((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)));
            case concat:
                final int l = e.getArg().get(2).getIntValue();
                return engine.concat((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)), l);
            case extract:
                final int left = e.getArg().get(1).getIntValue();
                final int right = e.getArg().get(2).getIntValue();
                return engine.extract((IntExpr) encodeElement(e.getArg().get(0)), left, right);
            case M:
                return engine.M((IntExpr) encodeElement(e.getArg().get(0)), (IntExpr) encodeElement(e.getArg().get(1)),(IntExpr) encodeElement(e.getArg().get(2)));
            default:
                LOGGER.error("Unknown expression!");
                return null;
        }

    }

    public void encode(ArrayList<EExpr> exprs) {
        for (EExpr e : exprs) {
            rules.add((BoolExpr) encodeElement(e));
        }
    }

    public BoolExpr encodeQuery(EExpr query) {
        return (BoolExpr) encodeElement(query);
    }
}